from typing import final

from dataspree.inference.core.config.configurable import Configurable

from dataspree.inference.core.signals import BaseSignal

import logging

logger = logging.getLogger(__name__)


class SignalHandler(Configurable):
    """ An instance of a signal handler subclass works as a node in the signal routing graph.
    It can receive signals and propagate them to others.
    The exact behavior is defined in the implementation of the class.
    """

    def __init__(self):
        Configurable.__init__(self)
        # Identifier used in order to check equality of objects derived from the instance that was deep or shallow
        # copied along the way
        self.__og_id: int = id(self)
        self.logger = logging.getLogger(__name__)

    @final
    def handle_signal(self, signal: BaseSignal) -> None:
        self.process_signal(signal)
        logger.debug(f'{self} received {signal}')
        signal.transmit(self)

    def propagate_signal(self, signal: BaseSignal) -> None:
        pass

    def process_signal(self, signal: BaseSignal) -> None:
        """
        @dev implementing @singledispatchmethod.process_signal here, and decorating the implementations
        with @SignalHandler.process_signal.register is not possible.
        This causes the signal dispatcher to be not tied to the instance.
        If there are multiple classes that register a dispatch method, this would cause all of them to be active
        whenever one of them is supposed to be invoked.

        :param signal:  the signal that is supposed to be dispatched.
        """
        self.logger.info(f'{self} received {signal}')

    def __eq__(self, other) -> bool:
        return isinstance(other, SignalHandler) and self.__og_id == other.__og_id
