from __future__ import annotations

import datetime
from typing import List, TypeVar, Generic, Any

_SourceT = TypeVar('_SourceT', bound='SignalHandler')
_T = TypeVar('_T')


def non_copyable(cls):
    """Decorator to make class instances only return themselves when copied or deep-copied."""
    cls.__copy__ = lambda self: self
    cls.__deepcopy__ = lambda self, memo: self
    return cls


@non_copyable
class SignalSource(Generic[_SourceT]):
    """A wrapper class for a signal source that should be used for any vertex in the signal routing graph.

    It is used to prevent copying the source instance when the signal is transmitted.
    """

    def __init__(self, source: _SourceT) -> None:
        self._source_instance = source
        self._timestamp = datetime.datetime.now()

    @property
    def source_instance(self) -> _SourceT:
        return self._source_instance

    @property
    def timestamp(self) -> datetime.datetime:
        return self._timestamp

    def __eq__(self, other: Any):
        return isinstance(other, SignalSource) and self._source_instance == other.source_instance

    def __str__(self):
        return str(self._source_instance)


class BaseSignal(Generic[_SourceT]):
    """Base class for all signals that can be transmitted. It does not require an initial source."""

    def __init__(self):
        self._routing: List[SignalSource[_SourceT]] = []
        self._signal_id: int = id(self)

    @property
    def signal_id(self) -> int:
        return self._signal_id

    def transmit(self, source: _SourceT):
        """Propagate the signal as defined in the implementation of the propagate_signal method of the SignalHandler.
        In any case, prevent cyclic transmission.
        """
        for node in self._routing:
            if source == node:
                return
        self._routing.append(SignalSource(source))
        source.propagate_signal(self)

    def __str__(self):
        route_str = '->'.join((str(r) for r in self._routing))
        return f'{type(self).__name__} ({self.signal_id}) via {route_str}'


class Signal(Generic[_SourceT], BaseSignal[_SourceT]):
    """A signal which is initially emitted by a source."""

    def __init__(self, source: _SourceT):
        BaseSignal.__init__(self)
        self._routing.append(SignalSource(source))

    @property
    def _source(self) -> _SourceT:
        return self._routing[0].source_instance


class NodeSignal(Signal['Node']):
    """A signal which is propagated via inference nodes (dataspree.inference.core.node.Node)."""

    @property
    def source_type(self):
        return type(self._source)

    @property
    def source_name(self):
        return self._source.name


class TerminatedNodeSignal(NodeSignal):
    pass
